# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from no_vtf.filesystem.input_paths import InputPaths
from no_vtf.filesystem.output_directories import OutputDirectories

__all__ = [
    "InputPaths",
    "OutputDirectories",
]
